"""Tests for delete CLI functions"""

import logging
from click import Context
from click.testing import CliRunner

from housekeeper.cli import delete


def test_delete_files_non_specified(base_context: Context, cli_runner: CliRunner, caplog):
    """Test to delete files without specifying bundle name or tag"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a store and a cli runner

    # WHEN trying to delete files without specifying bundle name or tag
    result = cli_runner.invoke(delete.files_cmd, [], obj=base_context)

    # THEN assert it exits non zero
    assert result.exit_code == 1
    # THEN HAL9000 should interfere
    assert "Please specify a bundle or a tag" in caplog.text


def test_delete_files_non_existing_bundle(
    base_context: Context, cli_runner: CliRunner, case_id: str, caplog
):
    """Test to delete a non existing bundle"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a store and a cli runner

    # WHEN trying to delete a bundle
    result = cli_runner.invoke(delete.files_cmd, ["--bundle-name", case_id], obj=base_context)

    # THEN assert it exits non zero
    assert result.exit_code == 1
    # THEN it should communicate that the bundle was not found
    assert "Bundle not found" in caplog.text


def test_delete_existing_bundle_with_confirmation(
    populated_context: Context, cli_runner: CliRunner, case_id: str, caplog
):
    """Test to delete an existing bundle with confirmation"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a populated store and a cli runner
    store = populated_context["store"]
    # GIVEN a existing bundle
    bundle_obj = store.Bundle.query.first()
    assert bundle_obj
    case_id = bundle_obj.name

    # WHEN trying to delete files without specifying bundle name or tag
    result = cli_runner.invoke(delete.files_cmd, ["--bundle-name", case_id], obj=populated_context)
    # THEN it should ask if you are sure
    assert "Are you sure you want to delete" in result.output


def test_delete_existing_bundle_no_confirmation(
    populated_context: Context, cli_runner: CliRunner, case_id: str, caplog
):
    """Test to delete an existing bundle without confirmation"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a populated store and a cli runner
    store = populated_context["store"]
    # GIVEN a existing bundle
    bundle_obj = store.Bundle.query.first()
    assert bundle_obj
    case_id = bundle_obj.name
    # GIVEN the bundle files
    query = store.files_before(bundle=case_id, tags=())
    nr_files = query.count()
    assert nr_files > 0

    # WHEN trying to delete a bundle without requiring confirmation
    cli_runner.invoke(delete.files_cmd, ["--bundle-name", case_id, "--yes"], obj=populated_context)
    # THEN the bundle should have been removed
    assert "deleted" in caplog.text



"""Tests for the store api"""
import datetime


def test_fetch_bundles(populated_store, bundle_data_old):
    """
    test fetch all files when two bundles are added
    """
    store = populated_store
    # GIVEN a store with two bundles and two files in each bundle
    bundle_old_obj, _ = store.add_bundle(data=bundle_data_old)
    store.add_commit(bundle_old_obj)

    # WHEN fetching all files in the database
    query = store.files_before(before="2020-05-04")

    # THEN all four files should be fetched
    assert len(query.all()) == 4


def test_fetch_past_files(populated_store, bundle_data_old, timestamp, old_timestamp):
    """
    test fetch files where not all files are older than before date
    """
    store = populated_store
    # GIVEN a store with two bundles and two files in each bundle
    bundle_old_obj, _ = store.add_bundle(data=bundle_data_old)
    store.add_commit(bundle_old_obj)

    # WHEN fetching all files before the oldest date
    date = old_timestamp + datetime.timedelta(days=10)
    assert old_timestamp < date < timestamp
    query = store.files_before(before=str(date))

    # THEN assert only files from the old bundle was found
    assert len(query.all()) == 2


def test_fetch_no_files_before_oldest(populated_store, bundle_data_old, old_timestamp, timestamp):
    """
    test fetch files where no files are older than before date
    """
    store = populated_store
    # GIVEN a store with two bundles and two files in each bundle
    bundle_old_obj, _ = store.add_bundle(data=bundle_data_old)
    store.add_commit(bundle_old_obj)

    # WHEN fetching all files before the oldest date
    date = old_timestamp - datetime.timedelta(days=10)
    assert date < old_timestamp < timestamp
    query = store.files_before(before=str(date))

    # THEN assert no files where that old
    assert len(query.all()) == 0



"""Base fixtures"""
import copy
import datetime
import json
import shutil
from copy import deepcopy
from pathlib import Path
from typing import List

import pytest
import yaml

from housekeeper.date import get_date
from housekeeper.store import Store, models

from .helper_functions import Helpers


# basic fixtures


@pytest.fixture(scope="function", name="vcf_tag_name")
def fixture_vcf_tag_name() -> str:
    """Return a tag named 'vcf'."""
    return "vcf"


@pytest.fixture(scope="function", name="family_tag_name")
def fixture_family_tag_name() -> str:
    """Return a tag named 'family'."""
    return "family"


@pytest.fixture(scope="function", name="sample_tag_name")
def fixture_sample_tag_name() -> str:
    """Return a tag named 'sample'."""
    return "sample"


@pytest.fixture(scope="function", name="non_existent_tag_name")
def fixture_non_existent_tag_name() -> str:
    """Return a non-existent tag name."""
    return "not_a_tag_name"


@pytest.fixture(scope="function", name="family_tag_names")
def fixture_family_tag_names(vcf_tag_name: str, family_tag_name: str) -> List[str]:
    """Return a list of the family tag names."""
    return [vcf_tag_name, family_tag_name]


@pytest.fixture(scope="function", name="sample_tag_names")
def fixture_sample_tag_names(vcf_tag_name: str, sample_tag_name: str) -> List[str]:
    """Return a list of the sample tag names."""
    return [vcf_tag_name, sample_tag_name]


@pytest.fixture(scope="function", name="case_id")
def fixture_case_id() -> str:
    """Return name of a case."""
    return "handsomepig"


@pytest.fixture(scope="function", name="other_case_id")
def fixture_other_case_id() -> str:
    """Return name of a another case."""
    return "maturecogar"


@pytest.fixture(scope="function", name="sample_data")
def fixture_sample_data(sample_tag_names: List[str], sample_vcf: Path) -> dict:
    """Return file and tags for sample."""
    return {"tags": sample_tag_names, "file": sample_vcf}


@pytest.fixture(scope="function", name="sample2_data")
def fixture_sample2_data(sample_tag_names: List[str], second_sample_vcf: Path) -> dict:
    """Return file and tags for sample."""
    return {"tags": sample_tag_names, "file": second_sample_vcf}


@pytest.fixture(scope="function", name="family_data")
def fixture_family_data(family_tag_names: List[str], family_vcf: Path) -> dict:
    """Return file and tags for sample."""
    return {"tags": family_tag_names, "file": family_vcf}


@pytest.fixture(scope="function", name="family2_data")
def fixture_family2_data(family_tag_names: List[str], second_family_vcf: Path) -> dict:
    """Return file and tags for sample."""
    return {"tags": family_tag_names, "file": second_family_vcf}


@pytest.fixture(scope="function", name="family3_data")
def fixture_family3_data(family_tag_names: List[str], third_family_vcf: Path) -> dict:
    """Return file and tags for sample."""
    return {"tags": family_tag_names, "file": third_family_vcf}


@pytest.fixture(scope="function", name="timestamp_string")
def fixture_timestamp_string() -> str:
    """Return a time stamp in str format."""
    return "2020-05-01"


@pytest.fixture(scope="function", name="timestamp")
def fixture_timestamp(timestamp_string: str) -> datetime.datetime:
    """Return a time stamp in date time format."""
    return get_date(timestamp_string)


@pytest.fixture(scope="function", name="later_timestamp")
def fixture_later_timestamp() -> datetime.datetime:
    """Return a time stamp in date time format to a later date."""
    return datetime.datetime(2020, 5, 25)


@pytest.fixture(scope="function", name="bundle_data")
def fixture_bundle_data(
        case_id: str, sample_data: dict, family_data: dict, timestamp: datetime.datetime
) -> dict:
    """Return a dummy bundle."""
    data = {
        "name": case_id,
        "created_at": timestamp,
        "files": [
            {
                "path": str(sample_data["file"]),
                "archive": False,
                "tags": sample_data["tags"],
            },
            {
                "path": str(family_data["file"]),
                "archive": True,
                "tags": family_data["tags"],
            },
        ],
    }
    return data


@pytest.fixture(scope="function", name="empty_version_data")
def fixture_empty_version_data(later_timestamp: datetime.datetime, case_id: str) -> dict:
    """Return a dummy bundle."""
    data = {"bundle_name": case_id, "created_at": later_timestamp, "files": []}
    return data


@pytest.fixture(scope="function", name="version_data")
def fixture_version_data(empty_version_data: dict, family2_data: dict, sample2_data: dict) -> dict:
    """Return a dummy bundle."""
    data = copy.deepcopy(empty_version_data)
    data["files"] = [
        {
            "path": str(sample2_data["file"]),
            "archive": False,
            "tags": sample2_data["tags"],
        },
        {
            "path": str(family2_data["file"]),
            "archive": True,
            "tags": family2_data["tags"],
        },
    ]
    return data


@pytest.fixture(scope="function", name="bundle_data_json")
def fixture_bundle_data_json(bundle_data: dict) -> str:
    """Return a dummy bundle."""
    json_data = copy.deepcopy(bundle_data)
    json_data["created_at"] = str(json_data.pop("created_at"))
    return json.dumps(json_data)


@pytest.fixture(scope="function", name="empty_version_data_json")
def fixture_empty_version_data_json(empty_version_data: dict) -> str:
    """Return a dummy bundle."""
    json_data = copy.deepcopy(empty_version_data)
    json_data["created_at"] = str(json_data.pop("created_at"))
    return json.dumps(json_data)


@pytest.fixture(scope="function", name="version_data_json")
def fixture_version_data_json(version_data: dict) -> str:
    """Return a dummy bundle."""
    json_data = copy.deepcopy(version_data)
    json_data["created_at"] = str(json_data.pop("created_at"))
    return json.dumps(json_data)


@pytest.fixture(scope="function", name="other_bundle")
def fixture_other_bundle(
        bundle_data: dict,
        other_case_id: str,
        later_timestamp: datetime.datetime,
        second_sample_vcf: Path,
        second_family_vcf: Path,
) -> dict:
    """Return a dummy bundle."""
    data = deepcopy(bundle_data)
    data["name"] = other_case_id
    data["created_at"] = later_timestamp
    data["files"][0]["path"] = str(second_sample_vcf)
    data["files"][1]["path"] = str(second_family_vcf)
    return data


@pytest.fixture(scope="function", name="db_name")
def fixture_db_name() -> str:
    """Return the name of a database."""
    return "hk_test.db"


@pytest.fixture(scope="function", name="db_path")
def fixture_db_path(db_dir: Path, db_name: str) -> Path:
    """Return the path to a database."""
    return db_dir / db_name


@pytest.fixture(scope="function", name="db_uri")
def fixture_db_uri(db_path: Path) -> str:
    """Return the uri to an in memory database."""
    return "sqlite:///" + str(db_path)


@pytest.fixture(scope="function", name="db_uri_memory")
def fixture_db_uri_memory() -> str:
    """Return the uri to an in memory database."""
    return "sqlite:///:memory:"


@pytest.fixture(scope="function", name="configs")
def fixture_configs(project_dir: Path, db_uri: str) -> dict:
    """Return a dict with housekeeper configs."""
    _configs = {"root": str(project_dir), "database": db_uri}
    return _configs


# object fixtures


@pytest.fixture(scope="function", name="vcf_tag_obj")
def fixture_vcf_tag_obj(vcf_tag_name: str, timestamp: datetime.datetime) -> str:
    """Return a tag object."""
    return models.Tag(name=vcf_tag_name, created_at=timestamp)


@pytest.fixture(scope="function", name="bundle_obj")
def fixture_bundle_obj(bundle_data: dict, store: Store) -> models.Bundle:
    """Return a bundle object."""
    return store.add_bundle(bundle_data)[0]


@pytest.fixture(scope="function", name="version_obj")
def fixture_version_obj(bundle_data: dict, store: Store) -> models.Version:
    """Return a version object."""
    return store.add_bundle(bundle_data)[1]


# dir fixtures


@pytest.fixture(scope="function", name="fixtures_dir")
def fixture_fixtures_dir() -> Path:
    """Return the path to the fixtures directory."""
    return Path("tests/fixtures/")


@pytest.fixture(scope="function", name="vcf_dir")
def fixture_vcf_dir(fixtures_dir: Path) -> Path:
    """Return the path to the vcf fixtures directory."""
    return fixtures_dir / "vcfs"


@pytest.fixture(scope="function", name="project_dir")
def fixture_project_dir(tmpdir_factory) -> Path:
    """Path to a temporary working directory."""
    my_tmpdir = Path(tmpdir_factory.mktemp("workdir"))
    yield my_tmpdir
    shutil.rmtree(str(my_tmpdir))


@pytest.fixture(scope="function", name="config_dir")
def fixture_config_dir(tmpdir_factory) -> Path:
    """Path to a temporary directory for config files."""
    my_tmpdir = Path(tmpdir_factory.mktemp("confdir"))
    yield my_tmpdir
    shutil.rmtree(str(my_tmpdir))


@pytest.fixture(scope="function", name="db_dir")
def fixture_db_dir(tmpdir_factory) -> Path:
    """Path to a temporary directory for databases."""
    my_tmpdir = Path(tmpdir_factory.mktemp("db_dir"))
    yield my_tmpdir
    shutil.rmtree(str(my_tmpdir))


# File fixtures


@pytest.fixture(scope="function", name="config_file")
def fixture_config_file(config_dir: Path, configs: dict) -> Path:
    """Create a config file and return the path to it."""
    conf_path = config_dir / "config.json"
    with open(conf_path, "w") as out_file:
        yaml.dump(configs, out_file)
    return conf_path


@pytest.fixture(scope="function", name="sample_vcf")
def fixture_sample_vcf(vcf_dir: Path) -> Path:
    """Return the path to a vcf file."""
    return vcf_dir / "example.vcf"


@pytest.fixture(scope="function", name="family_vcf")
def fixture_family_vcf(vcf_dir: Path) -> Path:
    """Return the path to a vcf file."""
    return vcf_dir / "family.vcf"


@pytest.fixture(scope="function", name="second_sample_vcf")
def fixture_second_sample_vcf(vcf_dir: Path) -> Path:
    """Return the path to a vcf file."""
    return vcf_dir / "example.2.vcf"


@pytest.fixture(scope="function", name="second_family_vcf")
def fixture_second_family_vcf(vcf_dir: Path) -> Path:
    """Return the path to a vcf file."""
    return vcf_dir / "family.2.vcf"


@pytest.fixture(scope="function", name="third_family_vcf")
def fixture_third_family_vcf(vcf_dir: Path) -> Path:
    """Return the path to a vcf file."""
    return vcf_dir / "family.3.vcf"


@pytest.fixture(scope="function", name="checksum_file")
def fixture_checksum_file(fixtures_dir: Path) -> Path:
    """Return the path to file to test checksum."""
    return fixtures_dir / "26a90105b99c05381328317f913e9509e373b64f.txt"


@pytest.fixture(scope="function", name="checksum")
def fixture_checksum(checksum_file: Path) -> Path:
    """Return the checksum for checksum test file."""
    return checksum_file.name.rstrip(".txt")


@pytest.fixture(scope="function", name="helpers")
def fixture_helpers() -> Helpers:
    """Return a test helper object."""
    return Helpers()


# Store fixtures


@pytest.fixture(scope="function", name="store")
def fixture_store(project_dir: Path) -> Store:
    """Return a store setup with all tables."""
    _store = Store(uri="sqlite:///", root=str(project_dir))
    _store.create_all()
    yield _store
    _store.drop_all()


@pytest.fixture(scope="function", name="populated_store")
def fixture_populated_store(store: Store, bundle_data: dict, helpers: Helpers) -> Store:
    """Returns a populated store."""
    helpers.add_bundle(store, bundle_data)
    return store



"""Fixtures for CLI tests"""
import datetime
import pytest
from click.testing import CliRunner
from copy import deepcopy
from pathlib import Path
from housekeeper.store import Store
from tests.helper_functions import Helpers


@pytest.fixture(scope="function", name="store")
def fixture_store(project_dir, db_uri):
    """Override the store fixture to get a controlled db path"""
    _store = Store(uri=db_uri, root=str(project_dir))
    _store.create_all()
    yield _store
    _store.drop_all()


@pytest.fixture(scope="function", name="cli_runner")
def fixture_cli_runner():
    """Return a cli runner for testing Click"""
    runner = CliRunner()
    return runner


@pytest.fixture(scope="function", name="base_context")
def fixture_base_context(db_uri, project_dir, store):
    """Return a context with initialized database"""
    _ctx = {
        "database": db_uri,
        "root": project_dir,
        "store": store,
    }
    return _ctx


@pytest.fixture(scope="function", name="populated_context")
def fixture_populated_context(db_uri, project_dir, populated_store):
    """Return a context with initialized database with some data"""
    _ctx = {
        "database": db_uri,
        "root": project_dir,
        "store": populated_store,
    }
    return _ctx


@pytest.fixture(scope="function", name="populated_store_subsequent")
def fixture_populated_store_subsequent(store: Store, bundle_data_subsequent: dict, helpers: Helpers) -> Store:
    """Returns a populated store"""
    helpers.add_bundle(store, bundle_data_subsequent)
    return store


@pytest.fixture(scope="function", name="populated_context_subsequent")
def fixture_populated_context_subsequent(db_uri, project_dir, populated_store_subsequent):
    """Return a context with initialized database with some data"""
    _ctx = {
        "database": db_uri,
        "root": project_dir,
        "store": populated_store_subsequent,
    }
    return _ctx


@pytest.fixture(scope="function", name="bundle_data_subsequent")
def fixture_bundle_data_subsequent(
        case_id: str, family_data: dict, family2_data: dict, family3_data: dict, timestamp: datetime.datetime
) -> dict:
    """Return a dummy bundle"""
    data = {
        "name": case_id,
        "created_at": timestamp,
        "files": [
            {
                "path": str(family_data["file"]),
                "archive": False,
                "tags": family_data["tags"],
            },
            {
                "path": str(family2_data["file"]),
                "archive": False,
                "tags": family2_data["tags"],
            },
            {
                "path": str(family3_data["file"]),
                "archive": True,
                "tags": family3_data["tags"],
            }
        ],
    }
    return data



"""Tests for finding tags in store."""
from housekeeper.store import Store
from housekeeper.store.models import Tag


def test_tag_with_tag_name(populated_store: Store, sample_tag_name: str):
    """Test fetching a tag from database given a tag name."""
    # GIVEN a populated store and a tag name

    # WHEN retrieving a tag from Store
    test_tag = populated_store.get_tag(tag_name=sample_tag_name)

    # THEN a tag should be returned
    assert isinstance(test_tag, Tag)

    # THEN the tag should have the tag name gotten as parameter
    assert test_tag.name == sample_tag_name


def test_tag_without_tag_name(populated_store: Store):
    """Test fetching a tag from database without specifying a tag name returns None."""
    # GIVEN a populated store

    # WHEN retrieving a tag from Store
    test_tag = populated_store.get_tag()

    # THEN a tag should be returned
    assert test_tag is None



"""Class with helper functions to use in tests"""

import json
from typing import Iterable, List

from housekeeper.constants import LOGLEVELS
from housekeeper.store import Store


class Helpers:
    """Hold small methods that might be helpful for the tests"""

    @staticmethod
    def count_iterable(iter_obj: Iterable) -> int:
        """Count the length of a iterable"""
        return sum(1 for item in iter_obj)

    @staticmethod
    def get_stdout(output: str) -> str:
        """Strip log messages from a string"""
        stripped = []
        # Strip logging lines
        for line in output.split("\n"):
            if any(log_level in line for log_level in LOGLEVELS):
                continue
            stripped.append(line)
        return "\n".join(stripped)

    @staticmethod
    def get_json(output: str) -> List:
        """Convert a string to json"""
        output = Helpers.get_stdout(output)
        return json.loads(output)

    @staticmethod
    def add_bundle(store: Store, bundle: dict) -> None:
        """Add and commit bundle to housekeeper store"""
        bundle_obj, _ = store.add_bundle(bundle)
        store.add_commit(bundle_obj)



"""Tests for delete version CLI functions"""
import logging
from click import Context
from click.testing import CliRunner

from housekeeper.cli import delete


def test_delete_non_version_no_input(base_context: Context, cli_runner: CliRunner, caplog):
    """Test to delete a version without any input

    The command should fail since there is no input
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a store and a cli runner

    # WHEN trying to delete a version without specifying anything
    result = cli_runner.invoke(delete.version_cmd, [], obj=base_context)

    # THEN assert it exits non zero
    assert result.exit_code == 1
    # THEN it should communicate that the bundle was not found
    assert "Please select a bundle or a version" in caplog.text


# def test_delete_existing_bundle_with_version(
#     populated_context: Context, cli_runner: CliRunner, case_id: str, caplog
# ):
#     """Test to delete an existing bundle with versions"""
#     caplog.set_level(logging.DEBUG)
#     # GIVEN a context with a store and a cli runner
#     store = populated_context["store"]
#     # GIVEN a bundle with versions
#     bundle_obj = store.bundle(name=case_id)
#     assert len(bundle_obj.versions) > 0
#
#     # WHEN trying to delete a bundle
#     result = cli_runner.invoke(delete.bundle_cmd, [case_id], obj=populated_context)
#
#     # THEN it should ask if you are sure
#     assert "Bundle has versions, can not delete bundle" in caplog.text
#
#
# def test_delete_existing_bundle_no_versions_no_confirmation(
#     populated_context: Context, cli_runner: CliRunner, case_id: str, caplog
# ):
#     """Test to delete an existing bundle without confirmation"""
#     caplog.set_level(logging.DEBUG)
#     # GIVEN a context with a store and a cli runner
#     store = populated_context["store"]
#     # GIVEN a bundle without versions
#     cli_runner.invoke(delete.version_cmd, ["-b", case_id], obj=populated_context, input="Yes")
#     bundle_obj = store.bundle(name=case_id)
#     assert len(bundle_obj.versions) == 0
#
#     # WHEN trying to delete a bundle
#     result = cli_runner.invoke(delete.bundle_cmd, [case_id], obj=populated_context, input="no")
#     # THEN assert it exits non zero
#     assert result.exit_code == 1
#
#
# def test_delete_existing_bundle_no_versions_with_confirmation(
#     populated_context: Context, cli_runner: CliRunner, case_id: str, caplog
# ):
#     """Test to delete an existing bundle without confirmation"""
#     caplog.set_level(logging.DEBUG)
#     # GIVEN a context with a store and a cli runner
#     store = populated_context["store"]
#     # GIVEN a bundle without versions
#     cli_runner.invoke(delete.version_cmd, ["-b", case_id], obj=populated_context, input="Yes")
#     bundle_obj = store.bundle(name=case_id)
#     assert len(bundle_obj.versions) == 0
#
#     # WHEN trying to delete a bundle
#     result = cli_runner.invoke(delete.bundle_cmd, [case_id], obj=populated_context, input="Yes")
#     # THEN assert it exits non zero
#     assert result.exit_code == 0
#     # THEN it should communicate that it was deleted
#     assert "Bundle deleted" in caplog.text



"""Tests for store core functions"""

from housekeeper.store import models

# tag tests


def test_create_tag_obj(store, vcf_tag_name):
    """Test to create a tag object"""
    # GIVEN a store and a tag name
    # WHEN creating a tag
    new_tag = store.new_tag(vcf_tag_name)
    # THEN assert a tag object was created
    assert isinstance(new_tag, models.Tag)
    assert new_tag.name == vcf_tag_name


def test_add_tag(store, vcf_tag_obj):
    """Test to add a tag"""
    # GIVEN a store without tags
    assert store.Tag.query.count() == 0
    # WHEN adding a tag
    store.add_commit(vcf_tag_obj)
    # THEN assert the new tag was added
    assert store.Tag.query.count() == 1
    assert store.Tag.query.first() == vcf_tag_obj


# version tests


def test_create_version_obj(store, timestamp):
    """Test to create a version object"""
    # GIVEN a store and a time stamp
    # WHEN creating a version
    new_version = store.new_version(created_at=timestamp, expires_at=timestamp)
    # THEN assert a version object was created
    assert isinstance(new_version, models.Version)
    assert new_version.created_at == timestamp


# bundle tests


def test_create_bundle_obj(store, bundle_data):
    """Test to create a bundle object"""
    # GIVEN some bundle information
    # WHEN adding the new bundle
    bundle_obj = store.add_bundle(bundle_data)[0]
    # THEN the bundle should have correct name
    assert bundle_obj.name == bundle_data["name"]
    # THEN the bundle should have correct creation time
    assert bundle_obj.created_at == bundle_data["created_at"]
    # THEN assert that a version was added
    assert len(bundle_obj.versions) == 1
    # THEN assert that all files where added to the bundle
    assert len(bundle_obj.versions[0].files) == len(bundle_data["files"])


def test_add_bundle(store, bundle_obj):
    """Test to add a bundle to the store"""
    # GIVEN a store without files, tags, versions or bundles
    assert store.Bundle.query.count() == 0
    assert store.Tag.query.count() == 0
    assert store.Version.query.count() == 0
    assert store.File.query.count() == 0
    # WHEN adding the new bundle
    store.add_commit(bundle_obj)
    # THEN assert that a bundle was added
    assert store.Bundle.query.count() == 1
    # THEN assert that the bundle is correct
    assert store.Bundle.query.first() == bundle_obj
    # THEN assert some tags where added
    assert store.Tag.query.count() > 0
    # THEN assert some files where added
    assert store.File.query.count() > 0
    # THEN assert we have a version
    assert store.Version.query.count() > 0


def test_add_bundle_twice(populated_store, bundle_data):
    """Test to add a bundle twice"""
    store = populated_store
    # GIVEN a store ppopulated with a bundle
    assert store.Bundle.query.count() > 0
    # WHEN adding the same bundle again
    new_bundle = store.add_bundle(bundle_data)
    # THEN it should return None
    assert new_bundle is None


def test_add_two_versions_of_bundle(populated_store, second_bundle_data):
    """Test to add two versions of the same bundle"""
    store = populated_store
    # GIVEN a populated store and some modified bundle data
    assert store.Bundle.query.count() > 0

    # WHEN adding the modified bundle to the database
    new_bundle_obj = store.add_bundle(second_bundle_data)[0]
    store.add_commit(new_bundle_obj)

    # THEN there should still be one bundle
    assert store.Bundle.query.count() == 1
    # THEN there should be two versions
    assert store.Version.query.count() == 2
    # THEN tere should be all four files
    assert store.File.query.count() == 4



"""Code for CLI get"""
import json as jsonlib
import logging
from typing import List

import click
from rich.console import Console

from housekeeper.store.api import schema

from .tables import get_bundles_table, get_files_table, get_tags_table, get_versions_table

LOG = logging.getLogger(__name__)


@click.group()
def get():
    """Get info from database"""


@get.command("bundle")
@click.argument("bundle-name", required=False)
@click.option("-i", "--bundle-id", type=int, help="Search for a bundle with bundle id")
@click.option("-j", "--json", is_flag=True, help="Output to json format")
@click.option("-v", "--verbose", is_flag=True, help="List files from latest version")
@click.option("-c", "--compact", is_flag=True, help="print compact filenames IFF verobe flag present")
@click.pass_context
def bundle_cmd(context, bundle_name, bundle_id, json, verbose, compact):
    """Get bundle information from database"""
    store = context.obj["store"]
    bundle_objs = store.bundles()
    if bundle_name or bundle_id:
        bundle_obj = store.bundle(name=bundle_name, bundle_id=bundle_id)
        bundle_objs = [bundle_obj] if bundle_obj else []
    if not bundle_objs:
        LOG.info("Could not find any bundles")
        return
    template = schema.BundleSchema()
    result = []
    for bundle_obj in bundle_objs:
        result.append(template.dump(bundle_obj))

    if json:
        click.echo(jsonlib.dumps(result, indent=4, sort_keys=True))
        return
    console = Console()
    console.print(get_bundles_table(result))
    if verbose:
        for bundle in bundle_objs:
            if len(bundle.versions) == 0:
                LOG.info("No versions found for bundle %s", bundle.name)
                return
            version_obj = bundle.versions[0]
            context.invoke(version_cmd, version_id=version_obj.id, verbose=True, compact=compact)


@get.command("version")
@click.option("-b", "--bundle-name", help="Fetch all versions from a bundle")
@click.option("-i", "--version-id", type=int, help="Fetch a specific version")
@click.option("-j", "--json", is_flag=True, help="Output to json format")
@click.option("-v", "--verbose", is_flag=True, help="print additional information")
@click.option("-c", "--compact", is_flag=True, help="print compact filenames IFF verobe flag present")
@click.pass_context
def version_cmd(context, bundle_name, json, version_id, verbose, compact):
    """Get versions from database"""
    store = context.obj["store"]
    if not (bundle_name or version_id):
        LOG.info("Please select a bundle or a version")
        return
    if bundle_name:
        bundle = store.bundle(name=bundle_name)
        if not bundle:
            LOG.info("Could not find bundle %s", bundle_name)
            return
        version_objs = bundle.versions

    if version_id:
        version = store.Version.get(version_id)
        if not version:
            LOG.warning("Could not find version %s", version_id)
            raise click.Abort
        version_objs = [version]

    version_template = schema.VersionSchema()
    result = []
    for version_obj in version_objs:
        bundle_obj = store.bundle(bundle_id=version_obj.bundle_id)
        res = version_template.dump(version_obj)
        res["bundle_name"] = bundle_obj.name
        result.append(res)

    if json:
        click.echo(jsonlib.dumps(result))
        return

    console = Console()
    console.print(get_versions_table(result))
    if not verbose:
        return

    for version_obj in version_objs:
        context.invoke(files_cmd, version=version_obj.id, verbose=True, compact=compact)


@get.command("file")
@click.option("-t", "--tag", "tags", multiple=True, help="filter by file tag")
@click.option("-v", "--version", type=int, help="filter by version of the bundle")
@click.option("-V", "--verbose", is_flag=True, help="print additional information")
@click.option("-j", "--json", is_flag=True, help="Output to json format")
@click.option("-c", "--compact", is_flag=True, help="print compact filenames")
@click.argument("bundle", required=False)
@click.pass_context
def files_cmd(context, tags: List[str], version: int, verbose: bool, bundle: str, json: bool, compact: bool):
    """Get files from database"""
    store = context.obj["store"]
    file_objs = store.files(bundle=bundle, tags=tags, version=version)
    template = schema.FileSchema()
    result = []
    for file_obj in file_objs:
        result.append(template.dump(file_obj))
    if json:
        click.echo(jsonlib.dumps(result))
        return
    console = Console()
    console.print(get_files_table(result, verbose=verbose, compact=compact))


@get.command("tag")
@click.option("-j", "--json", is_flag=True, help="Output to json format")
@click.option("-n", "--name", multiple=True, help="Specify a tag name")
@click.pass_context
def tag_cmd(context, json, name):
    """Get the tags from database"""
    store = context.obj["store"]
    LOG.info("Fetch tags")
    tag_objs = store.get_tags()
    template = schema.TagSchema()
    result = []
    for tag_obj in tag_objs:
        if name and (tag_obj.name not in name):
            continue
        LOG.debug("Use tag %s", tag_obj.name)
        result.append(template.dump(tag_obj))
    if not result:
        LOG.info("Could not find any of the specified tags [%s]", ", ".join(name))
        return
    if json:
        click.echo(jsonlib.dumps(result))
        return
    console = Console()
    console.print(get_tags_table(result))



"""Tests foe adding files via CLI"""
import logging
from pathlib import Path

from click import Context
from click.testing import CliRunner

from housekeeper.cli.add import file_cmd


def test_add_file_non_existing_bundle(
    base_context: Context,
    cli_runner: CliRunner,
    case_id: str,
    second_sample_vcf: Path,
    caplog,
):
    """Test to add a file to a non existing bundle"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a empty store and a cli runner
    unknown_bundle_name = case_id

    # WHEN trying to add a bundle
    result = cli_runner.invoke(
        file_cmd, [str(second_sample_vcf), "-b", unknown_bundle_name], obj=base_context
    )

    # THEN assert it fails
    assert result.exit_code == 1
    # THEN check that the proper information is displayed
    assert f"unknown bundle: {unknown_bundle_name}" in caplog.text


def test_add_file_json_and_file_path(
    base_context: Context,
    cli_runner: CliRunner,
    file_data_json: str,
    sample_vcf: Path,
    caplog,
):
    """Test to add a file to a non existing bundle"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a empty store and a cli runner

    # WHEN trying to add a bundle
    result = cli_runner.invoke(
        file_cmd, [str(sample_vcf), "--json", file_data_json], obj=base_context
    )

    # THEN assert it fails
    assert result.exit_code == 1
    # THEN check that the proper information is displayed
    assert "Can not input both json and path" in caplog.text


def test_add_file_no_input(base_context: Context, cli_runner: CliRunner, caplog):
    """Test to add a file without specifying any input"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a empty store and a cli runner

    # WHEN trying to add a bundle
    result = cli_runner.invoke(file_cmd, [], obj=base_context)

    # THEN assert it fails since there is no valid input
    assert result.exit_code == 1
    # THEN check that the proper information is displayed
    assert "Please input json or path" in caplog.text


def test_add_non_existing_file(base_context: Context, cli_runner: CliRunner, caplog):
    """Test to add a file without specifying any input"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a empty store and a cli runner
    non_existing = Path("hello.txt")
    assert non_existing.exists() is False

    # WHEN trying to add a bundle
    result = cli_runner.invoke(file_cmd, [str(non_existing)], obj=base_context)

    # THEN assert it fails since there is no valid input
    assert result.exit_code == 1
    # THEN check that the proper information is displayed
    assert f"File: {non_existing} does not exist" in caplog.text


def test_add_file_existing_bundle(
    populated_context: CliRunner,
    cli_runner: CliRunner,
    case_id: str,
    second_sample_vcf: Path,
    caplog,
):
    """Test to add a file to a existing bundle"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a populated store and a cli runner
    bundle_name = case_id

    # WHEN trying to add the file to a bundle
    result = cli_runner.invoke(
        file_cmd, [str(second_sample_vcf), "-b", bundle_name], obj=populated_context
    )

    # THEN assert it succedes
    assert result.exit_code == 0
    # THEN check that the proper information is displayed
    assert "new file added" in caplog.text


def test_add_file_json(
    populated_context: Context,
    cli_runner: CliRunner,
    file_data_json: str,
    caplog,
):
    """Test to add a file in json format to a non existing bundle"""
    caplog.set_level(logging.DEBUG)
    # GIVEN a populated store and a cli runner

    # WHEN trying to add a bundle
    result = cli_runner.invoke(file_cmd, ["--json", file_data_json], obj=populated_context)

    # THEN assert it fails
    assert result.exit_code == 0
    # THEN check that the proper information is displayed
    assert "new file added" in caplog.text



"""
This module handles finding things in the store/database
"""
import datetime as dt
import logging
from pathlib import Path
from typing import Iterable, List, Set

from sqlalchemy import func as sqlalchemy_func
from sqlalchemy.orm import Query

from housekeeper.date import get_date
from housekeeper.store.models import Bundle, File, Tag, Version
from housekeeper.store.bundle_filters import apply_bundle_filter, BundleFilters
from housekeeper.store.tag_filters import apply_tag_filter, TagFilter
from housekeeper.store.version_filters import apply_version_filter, VersionFilters
from housekeeper.store.version_bundle_filters import apply_version_bundle_filter, VersionBundleFilters

from .base import BaseHandler

LOG = logging.getLogger(__name__)


class FindHandler(BaseHandler):
    """Handler for searching the database"""

    def bundles(self):
        """Fetch bundles."""
        LOG.info("Fetching all bundles")
        return self._get_bundle_query()

    def _get_bundle_query(self) -> Query:
        """Return bundle query."""
        return self.Bundle.query

    def _get_version_query(self) -> Query:
        """Return version query."""
        return self.Version.query

    def _get_version_bundle_query(self) -> Query:
        """Return version bundle query."""
        return self.Version.query.join(Version.bundle)

    def bundle(self, name: str = None, bundle_id: int = None) -> Bundle:
        """Fetch a bundle from the store."""
        if bundle_id:
            LOG.info(f"Fetching bundle with id: {bundle_id}")
            return apply_bundle_filter(
                bundles=self._get_bundle_query(),
                filter_functions=[BundleFilters.FILTER_BY_ID],
                bundle_id=bundle_id,
            ).first()

        LOG.info(f"Fetching bundle with name: {name}")
        return apply_bundle_filter(
            bundles=self._get_bundle_query(),
            filter_functions=[BundleFilters.FILTER_BY_NAME],
            bundle_name=name,
        ).first()

    def version(
        self, bundle: str = None, date: dt.datetime = None, version_id: int = None
    ) -> Version:
        """Fetch a version from the store."""
        if version_id:
            LOG.info(f"Fetching version with id: {version_id}")
            return apply_version_filter(
                versions=self._get_version_query(),
                filter_functions=[VersionFilters.FILTER_BY_ID],
                version_id=version_id,
            ).first()

        return apply_version_bundle_filter(
            version_bundles=self._get_version_bundle_query(),
            filter_functions=[VersionBundleFilters.FILTER_BY_DATE_AND_NAME],
            version_date=date,
            bundle_name=bundle,
        ).first()

    def get_tag(self, tag_name: str = None) -> Tag:
        """Return a tag from the database."""
        LOG.info(f"Fetching tag with name: {tag_name}")
        return apply_tag_filter(
            tags=self._get_tag_query(),
            filter_functions=[TagFilter.FILTER_BY_NAME],
            tag_name=tag_name,
        ).first()

    def get_tags(self) -> Query:
        """Return all tags from the database."""
        LOG.info("Fetching all tags")
        return self._get_tag_query()

    def _get_tag_query(self) -> Query:
        """Return a tag query."""
        return self.Tag.query

    def file_(self, file_id: int) -> File:
        """Returns a file by record id."""
        return self.File.get(file_id)

    def files(
        self, *, bundle: str = None, tags: List[str] = None, version: int = None, path: str = None
    ) -> Iterable[File]:
        """Fetch files from the store."""
        query = self.File.query
        if bundle:
            LOG.info(f"Fetching files from bundle {bundle}")
            query = query.join(self.File.version, self.Version.bundle).filter(
                self.Bundle.name == bundle
            )

        if tags:
            formatted_tags = ",".join(tags)
            LOG.info(f"Fetching files with tags in [{formatted_tags}]")
            # require records to match ALL tags
            query = (
                query.join(self.File.tags)
                .filter(self.Tag.name.in_(tags))
                .group_by(File.id)
                .having(sqlalchemy_func.count(Tag.name) == len(tags))
            )

        if version:
            LOG.info(f"Fetching files from version {version}")
            query = query.join(self.File.version).filter(self.Version.id == version)

        if path:
            LOG.info(f"Fetching files with path {path}")
            query = query.filter_by(path=path)

        return query

    def files_before(
        self, *, bundle: str = None, tags: List[str] = None, before: str = None
    ) -> File:
        """Fetch files before date from store"""
        query = self.files(tags=tags, bundle=bundle)
        if before:
            try:
                before_dt = get_date(before)
            except ValueError:
                before_dt = get_date(before, "%Y-%m-%d %H:%M:%S")
            query = query.join(Version).filter(Version.created_at < before_dt)

        return query

    @staticmethod
    def files_ondisk(file_objs: File) -> Set[File]:
        """Returns a list of files that are on disk."""

        files_on_disk = {file_obj for file_obj in file_objs if Path(file_obj.full_path).is_file()}
        return files_on_disk



"""Tests for the add files method in store"""

from housekeeper.store import Store
from housekeeper.store.models import File, Tag, Bundle

from pathlib import Path
from typing import List


def test_add_file(populated_store: Store, second_family_vcf: Path, family_tag_names: List[str]):
    """Test to create a file with the add file method"""
    # GIVEN the path and the tags for a file

    # GIVEN a store populated with a bundle
    bundle: Bundle = populated_store.bundles().first()
    assert isinstance(bundle, Bundle)

    # WHEN using the add file method to create a new file object
    new_file: File = populated_store.add_file(
        file_path=second_family_vcf, bundle=bundle, tags=family_tag_names
    )

    # THEN assert that the file is a file object
    assert isinstance(new_file, File)
    # THEN assert that the file is added to the latest version of the bundle
    assert new_file.version == bundle.versions[0]
    # THEN assert that the tags are added to the new file
    assert len(new_file.tags) == len(family_tag_names)
    for tag_obj in new_file.tags:
        assert isinstance(tag_obj, Tag)


def test_add_file_no_tags(populated_store: Store, second_family_vcf: Path):
    """Test to create a file with the add file method without tags"""
    # GIVEN a path for a file

    # GIVEN a store populated with a bundle
    bundle: Bundle = populated_store.bundles().first()
    assert isinstance(bundle, Bundle)

    # WHEN using the add file method to create a new file object
    new_file = populated_store.add_file(file_path=second_family_vcf, bundle=bundle)

    # THEN assert that the no tags where added to the file
    assert len(new_file.tags) == 0



"""Tests for include cli command"""
import logging
from datetime import datetime
from click import Context
from click.testing import CliRunner

from housekeeper.cli.include import include
from housekeeper.cli.add import bundle_cmd


def test_include_files_creates_bundle_dir(populated_context: Context, cli_runner: CliRunner):
    """Test to include the files of a version for a bundle_name

    The bundle should not exist before and after command is run it should have been created
    """
    # GIVEN a context that is populated
    store = populated_context["store"]
    bundle_obj = store.Bundle.query.first()
    bundle_name = bundle_obj.name
    # GIVEN that the latest version of the bundle is not included
    assert bundle_obj.versions[0].included_at is None
    # GIVEN that no folder has been created since case is not included
    bundle_path = populated_context["root"] / bundle_obj.name
    assert not bundle_path.exists()

    # WHEN running the include files command
    result = cli_runner.invoke(include, [bundle_name], obj=populated_context)

    # THEN assert that the bundle path was created
    assert bundle_path.exists()


def test_include_files_creates_version_specific_bundle_dir(
    populated_context: Context, cli_runner: CliRunner
):
    """Test to include the files of a version for a bundle_name

    The version bundle should not exist before and after command is run it should have been created
    """
    # GIVEN a context that is populated
    store = populated_context["store"]
    bundle_obj = store.Bundle.query.first()
    version_obj = bundle_obj.versions[0]
    bundle_name = bundle_obj.name
    # GIVEN that the latest version of the bundle is not included
    assert version_obj.included_at is None
    # GIVEN that no version specific folder has been created since version is not included
    version_path = populated_context["root"] / bundle_obj.name / str(version_obj.created_at.date())
    assert not version_path.exists()

    # WHEN running the include files command
    result = cli_runner.invoke(include, [bundle_name], obj=populated_context)

    # THEN assert that the version path was created
    assert version_path.exists()


def test_include_files_adds_version_specific_files(
    populated_context: Context, cli_runner: CliRunner
):
    """Test to include the files of a version for a bundle_name

    The files should have been hard linked after command has been run
    """
    # GIVEN a context that is populated
    store = populated_context["store"]
    bundle_obj = store.Bundle.query.first()
    version_obj = bundle_obj.versions[0]
    bundle_name = bundle_obj.name
    # GIVEN that the latest version of the bundle is not included
    assert version_obj.included_at is None
    # GIVEN that no version specific folder has been created since version is not included
    version_path = populated_context["root"] / bundle_obj.name / str(version_obj.created_at.date())
    assert not version_path.exists()

    # WHEN running the include files command
    result = cli_runner.invoke(include, [bundle_name], obj=populated_context)

    # THEN assert that files are included in the folder
    files_included = []
    for file_path in version_path.iterdir():
        files_included.append(file_path)

    assert files_included
    assert len(files_included) == len(version_obj.files)


def test_include_files_specific_version(populated_context: Context, cli_runner: CliRunner):
    """Test to include the files of a version by specifying the version id

    The folder should have been created
    """
    # GIVEN a context that is populated
    store = populated_context["store"]
    bundle_obj = store.Bundle.query.first()
    version_obj = bundle_obj.versions[0]
    version_id = version_obj.id
    # GIVEN that the latest version of the bundle is not included
    assert version_obj.included_at is None
    # GIVEN that no version specific folder has been created since version is not included
    version_path = populated_context["root"] / bundle_obj.name / str(version_obj.created_at.date())
    assert not version_path.exists()

    # WHEN running the include files command
    result = cli_runner.invoke(include, ["--version-id", version_id], obj=populated_context)

    # THEN assert that the folder was created
    assert version_path.exists()

    # THEN assert that files are included in the folder
    files_included = []
    for file_path in version_path.iterdir():
        files_included.append(file_path)

    assert files_included
    assert len(files_included) == len(version_obj.files)


def test_include_version_no_args(populated_context: Context, cli_runner: CliRunner, caplog):
    """Test to include the files of a version without specifying bundle name or version

    The command should exit since it needs to be specified
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a context that is populated

    # WHEN running the include files command without args
    result = cli_runner.invoke(include, [], obj=populated_context)

    # THEN assert that the command exists with zero code 1
    assert result.exit_code == 1
    # THEN assert it was communicated that arguments are needed
    assert "use bundle name or version-id" in caplog.text


def test_include_non_existing_version(populated_context: Context, cli_runner: CliRunner, caplog):
    """Test to include the files of a version that does not exist

    The command should exit since a valid version is needed
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a context that is populated
    store = populated_context["store"]
    # GIVEN a version that does not exists
    version_id = 10
    assert not store.Version.get(version_id)

    # WHEN running the include files specifying the non existing version
    result = cli_runner.invoke(include, ["--version-id", version_id], obj=populated_context)

    # THEN assert that the command exists with zero code 1
    assert result.exit_code == 1
    # THEN assert it was communicated that version did not exist
    assert "version not found" in caplog.text


def test_include_non_existing_bundle(
    base_context: Context, cli_runner: CliRunner, timestamp: datetime, caplog
):
    """Test to include the files in the latest version of a bundle that does not exist

    The command should exit since a valid bundle is needed
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN context and a bundle_name that does not exist
    store = base_context["store"]
    bundle_name = "hello"
    assert not store.bundle(name=bundle_name)

    # WHEN running the include files specifying bundle without versions
    result = cli_runner.invoke(include, [bundle_name], obj=base_context)

    # THEN assert that the command exists with zero code 1
    assert result.exit_code == 1
    # THEN assert it was communicated that the bundle did not exist
    assert f"bundle {bundle_name} not found" in caplog.text


def test_include_bundle_without_version(
    base_context: Context, cli_runner: CliRunner, timestamp: datetime, caplog
):
    """Test to include the files in the latest version of a bundle that does not have versions

    The command should exit since a valid bundle with versions are needed
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN context with a bundle without versions
    store = base_context["store"]
    bundle_name = "hello"

    new_bundle = store.new_bundle(name=bundle_name, created_at=timestamp)
    store.add_commit(new_bundle)

    bundle_obj = store.bundle(name=bundle_name)
    assert len(bundle_obj.versions) == 0

    # WHEN running the include files specifying bundle without versions
    result = cli_runner.invoke(include, [bundle_name], obj=base_context)

    # THEN assert that the command exists with zero code 1
    assert result.exit_code == 1
    # THEN assert it was communicated that version did not exist
    assert f"Could not find any versions for bundle {bundle_name}" in caplog.text


def test_include_version_already_included(
    populated_context: Context, cli_runner: CliRunner, caplog
):
    """Test to include the files of a version that is already included

    The command should exit since the version can not be included again
    """
    caplog.set_level(logging.DEBUG)
    # GIVEN a context that is populated and a version that is already included
    store = populated_context["store"]
    version_obj = store.Version.query.first()
    version_id = version_obj.id
    result = cli_runner.invoke(include, ["--version-id", version_id], obj=populated_context)

    # WHEN trying to include the version again
    result = cli_runner.invoke(include, ["--version-id", version_id], obj=populated_context)

    # THEN assert that the command exists with zero code 1
    assert result.exit_code == 1
    # THEN assert it was communicated that version did not exist
    assert "version included on" in caplog.text



"""Module for deleting via CLI"""

import logging
import shutil
from pathlib import Path

import click

LOG = logging.getLogger(__name__)


@click.group()
def delete():
    """Delete things in the database."""


@delete.command("bundle")
@click.option("-y", "--yes", is_flag=True, help="skip checks")
@click.argument("bundle_name")
@click.pass_context
def bundle_cmd(context, yes, bundle_name):
    """Delete a empty bundle, that is a bundle without versions"""
    store = context.obj["store"]
    bundle_obj = store.bundle(bundle_name)
    if bundle_obj is None:
        LOG.warning("bundle %s not found", bundle_name)
        raise click.Abort

    if bundle_obj.versions:
        LOG.warning("Can not delete bundle, please remove all versions first")
        raise click.Abort

    question = f"Remove bundle {bundle_obj.name} from database?"
    if not (yes or click.confirm(question)):
        raise click.Abort

    bundle_obj.delete()
    store.commit()
    LOG.info("Bundle deleted: %s", bundle_obj.name)


@delete.command("version")
@click.option("-b", "--bundle-name", help="Fetch all versions from a bundle")
@click.option("-i", "--version-id", type=int, help="Fetch a specific version")
@click.option("-y", "--yes", is_flag=True, help="skip checks")
@click.pass_context
def version_cmd(context, bundle_name, version_id, yes):
    """Delete a version from database"""
    store = context.obj["store"]
    if not (bundle_name or version_id):
        LOG.info("Please select a bundle or a version")
        raise click.Abort

    if bundle_name:
        bundle_obj = store.bundle(name=bundle_name)
        if not bundle_obj:
            LOG.info("Could not find bundle %s", bundle_name)
            return
        if len(bundle_obj.versions) == 0:
            LOG.warning("Could not find versions for bundle %s", bundle_name)
            return
        LOG.info("Deleting the latest version of bundle %s", bundle_name)
        version_obj = bundle_obj.versions[0]

    if version_id:
        version = store.version(version_id=version_id)
        if not version:
            LOG.warning("Could not find version %s", version_id)
            raise click.Abort
        bundle_obj = store.bundle(bundle_id=version.bundle_id)
        for ver in bundle_obj.versions:
            if ver.id == version_id:
                version_obj = ver

    if version_obj.included_at:
        question = f"remove bundle version from file system and database: {version_obj.full_path}"
    else:
        question = f"remove bundle version from database: {version_obj.created_at.date()}"

    if not (yes or click.confirm(question)):
        raise click.Abort

    if version_obj.included_at:
        shutil.rmtree(version_obj.full_path, ignore_errors=True)

    version_obj.delete()
    store.commit()
    LOG.info("version deleted: %s", version_obj.full_path)


@delete.command("files")
@click.option("--yes", is_flag=True, help="skip checks")
@click.option("-t", "--tag", multiple=True, help="file tag")
@click.option("-b", "--bundle-name", help="bundle name")
@click.option("-a", "--before", help="version created before...")
@click.option("-n", "--notondisk", is_flag=True, help="rm db entry from files not on disk")
@click.option("-l", "--list-files", is_flag=True, help="lists files that will be deleted")
@click.option("-L", "--list-files-verbose", is_flag=True, help="lists additional information")
@click.pass_context
def files_cmd(context, yes, tag, bundle_name, before, notondisk, list_files, list_files_verbose):
    """Delete files based on tags."""
    store = context.obj["store"]
    file_objs = []
    if not (tag or bundle_name):
        LOG.info("Please specify a bundle or a tag")
        raise click.Abort

    if bundle_name:
        bundle_obj = store.bundle(bundle_name)
        if bundle_obj is None:
            LOG.warning("Bundle not found")
            raise click.Abort

    query = store.files_before(bundle=bundle_name, tags=tag, before=before)

    if notondisk:
        file_objs = set(query) - store.files_ondisk(query)
    else:
        file_objs = query.all()

    nr_files = 0
    for nr_files, file_obj in enumerate(file_objs, 1):
        if list_files_verbose:
            tags = ", ".join(tag.name for tag in file_obj.tags)
            click.echo(
                f"{click.style(str(file_obj.id), fg='blue')} | {file_obj.full_path} | "
                f"{click.style(tags, fg='yellow')}"
            )
        elif list_files:
            click.echo(file_obj.full_path)
        else:
            continue

    if nr_files == 0:
        LOG.warning("no files found")
        raise click.Abort

    if not (yes or click.confirm(f"Are you sure you want to delete {len(file_objs)} files?")):
        raise click.Abort

    for file_obj in file_objs:
        if yes or click.confirm(f"remove file from disk and database: {file_obj.full_path}"):
            file_obj_path = Path(file_obj.full_path)
            if file_obj.is_included and (file_obj_path.exists() or file_obj_path.is_symlink()):
                file_obj_path.unlink()
            file_obj.delete()
            store.commit()
            LOG.info("%s deleted", file_obj.full_path)


@delete.command("file")
@click.option("-y", "--yes", is_flag=True, help="skip checks")
@click.argument("file_id", type=int)
@click.pass_context
def file_cmd(context, yes, file_id):
    """Delete a file."""
    store = context.obj["store"]
    file_obj = store.File.get(file_id)
    if not file_obj:
        LOG.info("file not found")
        raise click.Abort

    if file_obj.is_included:
        question = f"remove file from file system and database: {file_obj.full_path}"
    else:
        question = f"remove file from database: {file_obj.full_path}"

    if yes or click.confirm(question):
        if file_obj.is_included and Path(file_obj.full_path).exists():
            Path(file_obj.full_path).unlink()

        file_obj.delete()
        store.commit()
        LOG.info("file deleted")
